r"""
======
Models
======
.. currentmodule:: epitome.models

.. autosummary::
  :toctree: _generate/

  VariationalPeakModel
  VLP
"""

import warnings

from epitome import GET_DATA_PATH, POSITIONS_FILE
with warnings.catch_warnings():
    warnings.filterwarnings("ignore",category=FutureWarning)
    import tensorflow as tf
    import tensorflow_probability as tfp
    from .functions import *
    
from .constants import *
from .generators import *
from .metrics import *
import numpy as np

import tqdm

with warnings.catch_warnings():
    warnings.filterwarnings("ignore",category=FutureWarning)
    import tensorflow as tf
    import tensorflow_probability as tfp

# for saving model
import pickle
from operator import itemgetter


#######################################################################
#################### Variational Peak Model ###########################
#######################################################################



class VariationalPeakModel():
    """ Model for learning from ChIP-seq peaks.
    """

    def __init__(self,
                 assays,
                 test_celltypes = [],
                 matrix = None,
                 assaymap = None,
                 cellmap = None,
                 debug = False,
                 batch_size = 64,
                 shuffle_size = 10,
                 prefetch_size = 10,
                 l1=0.,
                 l2=0.,
                 lr=1e-3,
                 radii=[1,3,10,30],
                 train_indices = None,
                 data = None,
                 data_path = None):
        """
        Initializes Peak Model

        Args:
            :param assays: list of assays to train model on
            :param test_celltypes: list of cell types to hold out for test. Should be in cellmap
            :param matrix: numpy matrix of indices mapping assay and cell to index in data
            :param assaymap: map of assays mapping assay name to row in matrix
            :param cellmap: map of cell types mapping cell name to column in matrix
            :param debug: used to print out intermediate validation values
            :param batch_size: batch size (default is 64)
            :param shuffle_size: data shuffle size (default is 10)
            :param prefetch_size: data prefetch size (default is 10)
            :param l1: l1 regularization (default is 0)
            :param l2: l2 regularization (default is 0)
            :param lr: lr (default is 1e-3)
            :param radii: radius of DNase-seq to consider around a peak of interest (default is [1,3,10,30])
            :param train_indices: option numpy array of indices to train from data[Dataset.TRAIN]
            :param data: data loaded from datapath. This option is mostly for testing, so users dont have to load in data for 
            :param data_path: path to data. Directory should contain all.pos.bed.gz, feature_name,test.npz,train.npz,valid.npz
            each model.
        """

        tf.compat.v1.logging.set_verbosity(tf.compat.v1.logging.INFO)
                    
        # user can provide their own assaymap information. 
        if assaymap is not None:
            assert matrix is not None and cellmap is not None, "matrix, cellmap, and assaymap must all be set"
        if cellmap is not None:
            assert matrix is not None and assaymap is not None, "matrix, cellmap, and assaymap must all be set"
        if matrix is not None:
            assert assaymap is not None and cellmap is not None, "matrix, cellmap, and assaymap must all be set"
            
        # get cell lines to train on if not specified
        if assaymap is None:
            # get list of TFs that have minimum number of cell lines
            matrix, cellmap, assaymap = get_assays_from_feature_file(eligible_assays = assays)
            assert len(assays) == len(list(assaymap))-1 
                   

        assert (set(test_celltypes) < set(list(cellmap))), \
                "test_celltypes %s must be subsets of available cell types %s" % (str(test_celltypes), str(list(cellmap)))

        # get evaluation cell types by removing any cell types that would be used in test
        self.eval_cell_types = list(cellmap).copy()
        self.test_celltypes = test_celltypes
        [self.eval_cell_types.remove(test_cell) for test_cell in self.test_celltypes]

        # load in data, if the user has not specified it
        if data is not None:
            self.data = data
        else:
            self.data = load_epitome_data()

        if not data_path:
            data_path = DATA_PATH

        self.regionsFile = os.path.join(data_path, POSITIONS_FILE)

        self.output_shape, self.train_iter = generator_to_tf_dataset(load_data(self.data[Dataset.TRAIN],
                                                self.eval_cell_types,
                                                self.eval_cell_types,
                                                matrix,
                                                assaymap,
                                                cellmap,
                                                radii = radii, mode = Dataset.TRAIN),
                                                batch_size, shuffle_size, prefetch_size)

        _,            self.valid_iter = generator_to_tf_dataset(load_data(self.data[Dataset.VALID],
                                                self.eval_cell_types,
                                                self.eval_cell_types,
                                                matrix,
                                                assaymap,
                                                cellmap,
                                                radii = radii, mode = Dataset.VALID),
                                                batch_size, 1, prefetch_size)

        # can be empty if len(test_celltypes) == 0
        _,            self.test_iter = generator_to_tf_dataset(load_data(self.data[Dataset.TEST],
                                               self.test_celltypes,
                                               self.eval_cell_types,
                                               matrix,
                                               assaymap,
                                               cellmap,
                                               radii = radii, mode = Dataset.TEST),
                                               batch_size, 1, prefetch_size)

        self.num_outputs = self.output_shape[0]
        self.l1, self.l2 = l1, l2
        self.lr = lr
        self.batch_size = batch_size
        self.prefetch_size = prefetch_size
        self.shuffle_size = shuffle_size
        self.optimizer =tf.compat.v1.train.AdamOptimizer(learning_rate=self.lr)

        # set self
        self.radii = radii
        self.debug = debug
        self.assaymap = assaymap
        self.test_celltypes = test_celltypes
        self.matrix = matrix
        self.assaymap= assaymap
        self.cellmap = cellmap
        self.model = self.create_model()

    def get_weight_parameters(self):
        """
        Extracts weight posterior statistics for layers with weight distributions.
        :param model: keras model

        :return triple of layer names, weight means for each layer and stddev for each layer.
        """

        names = []
        qmeans = []
        qstds = []
        for i, layer in enumerate(self.model.layers):
            try:
                q = layer.kernel_posterior
            except AttributeError:
                continue
            names.append("Layer {}".format(i))
            qmeans.append(q.mean())
            qstds.append(q.stddev())

        return (names, qmeans, qstds)

    def save(self, checkpoint_path):
        """
        Saves model.

        :param checkpoint_path: string file path to save model to.
        """
        # save keras model
        self.model.save(checkpoint_path)
        
        # save model params to pickle file
        dict_ = {'assays': list(self.assaymap),
                         'test_celltypes':self.test_celltypes,
                         'matrix':self.matrix,
                         'assaymap':self.assaymap,
                         'cellmap':self.cellmap,
                         'debug': self.debug,
                         'batch_size':self.batch_size,
                         'shuffle_size':self.shuffle_size,
                         'prefetch_size':self.prefetch_size,
                         'radii':self.radii}

        fileObject = open(os.path.join(checkpoint_path, "model_params.pickle"),'wb')
        pickle.dump(dict_,fileObject)
        fileObject.close()

    def body_fn(self):
        raise NotImplementedError()

    def g(self, p, a=1, B=0, y=1):
        """ Normalization Function. Normalizes loss w.r.t. label proportion.

        Constraints:
         1. g(p) = 1 when p = 1
         2. g(p) = a * p^y + B, where a, y and B are hyperparameters
        """
        return a * tf.math.pow(p, y) + B

    def loss_fn(self, y_true, y_pred, weights):
        """ Epitome's loss function. Sigmoid cross entropy weighted by missing labels.

        Args:
            :param y_true: vector of true values (0/1)
            :param y_pred: vector of predicted values
            :param weights: vector of 0/1 weights that negate missing truth

        Returns:
            sigmoid cross entropy loss

        """
        # weighted sum of cross entropy for non 0 weights
        # Reduction method = Reduction.SUM_BY_NONZERO_WEIGHTS
        loss = tf.compat.v1.losses.sigmoid_cross_entropy(y_true,
                                                        y_pred[:, Features.FEATURE_IDX.value, :],
                                                        weights = weights,
                                                        reduction = tf.compat.v1.losses.Reduction.SUM_BY_NONZERO_WEIGHTS)

        C = (len(self.assaymap)-1)
        # p = tf.math.reduce_sum(weights, axis=1)/C # should be of dimension 1 by batch size
        p = 1.0 # this is just taking the mean loss, nothing special here.
        return self.g(p)/C * loss


    def train(self, num_steps, lr=None, checkpoint_path = None):
        """ Trains an epitome model for specified number of steps.

        Args:
            :param num_steps: number of iterations to train.
            :param lr
            :param checkpoint_path: model path to continue training from

        """
        if lr == None:
            lr = self.lr

        tf.compat.v1.logging.info("Starting Training")

        @tf.function
        def train_step(inputs, labels, weights):
            with tf.GradientTape() as tape:
                logits = self.model(inputs, training=True)
                kl_loss = sum(self.model.losses)
                neg_log_likelihood = self.loss_fn(labels, logits, weights)
                elbo_loss = neg_log_likelihood + kl_loss

            gradients = tape.gradient(elbo_loss, self.model.trainable_variables)
            self.optimizer.apply_gradients(zip(gradients, self.model.trainable_variables))

            return elbo_loss, neg_log_likelihood, kl_loss

        for step, (inputs, labels, weights) in enumerate(self.train_iter.take(num_steps)):

            loss = train_step(inputs, labels, weights)

            if step % 1000 == 0:

                tf.compat.v1.logging.info(str(step) + " " + str(tf.reduce_mean(loss[0])) +
                                          str(tf.reduce_mean(loss[1])) +
                                          str(tf.reduce_mean(loss[2])))

                if (self.debug):
                    tf.compat.v1.logging.info("On validation")
                    _, _, _, _, _ = self.test(40000, log=False)
                    tf.compat.v1.logging.info("")

    def test(self, num_samples, mode = Dataset.VALID, calculate_metrics=False):
        """
        Tests model on valid and test dataset handlers.
        """

        if (mode == Dataset.VALID):
            handle = self.valid_iter # for standard validation of validation cell types

        elif (mode == Dataset.TEST and len(self.test_celltypes) > 0):
            handle = self.test_iter # for standard validation of validation cell types
        else:
            raise Exception("No data exists for %s. Use function test_from_generator() if you want to create a new iterator." % (mode))

        return self.run_predictions(num_samples, handle, calculate_metrics)

    def test_from_generator(self, num_samples, ds, calculate_metrics=True):
        """
        Runs test given a specified data generator
        :param num_samples: number of samples to test
        :param ds: tensorflow dataset, created by dataset_to_tf_dataset
        :param cell_type: cell type to test on. Used to generate holdout indices.

        :return predictions
        """
        return self.run_predictions(num_samples, ds, calculate_metrics)

    def eval_vector(self, data, vector, indices):
        """
        Evaluates a new cell type based on its chromatin (DNase or ATAC-seq) vector. len(vector) should equal
        the data.shape[1]

        Args:
            :param data: data to build features from
            :param vector: vector of 0s/1s of binding sites TODO AM 4/3/2019: try peak strength instead of 0s/1s
            :param indices: indices of vector to actually score. You need all of the locations for the generator.

            :return predictions for all factors
        """

        _,  ds = generator_to_tf_dataset(load_data(data,
                 self.test_celltypes,   # used for labels. Should be all for train/eval and subset for test
                 self.eval_cell_types,   # used for rotating features. Should be all - test for train/eval
                 self.matrix,
                 self.assaymap,
                 self.cellmap,
                 radii = self.radii,
                 mode = Dataset.RUNTIME,
                 dnase_vector = vector, indices = indices), self.batch_size, 1, self.prefetch_size)

        num_samples = len(indices)

        results = self.run_predictions(num_samples, ds, calculate_metrics = False)

        return results['preds_mean'], results['preds_std']

    def run_predictions(self, num_samples, iter_, calculate_metrics = True, samples = 50):
        """
        Runs predictions on num_samples records
        :param num_samples: number of samples to test
        :param iter_: output of self.sess.run(generator_to_one_shot_iterator()), handle to one shot iterator of records
        :param log: if true, logs individual factor accuracies

        :return preds, truth, assay_dict, auROC, auPRC, False
            preds = predictions,
            truth = actual values,
            sample_weight: 0/1 weights on predictions.
            assay_dict = if log=True, holds predictions for individual factors
            auROC = average macro area under ROC for all factors with truth values
            auPRC = average area under PRC for all factors with truth values
        """

        inv_assaymap = {v: k for k, v in self.assaymap.items()}

        batches = int(num_samples / self.batch_size)+1

        # empty arrays for concatenation
        truth = []
        preds_mean = []
        preds_std = []
        sample_weight = []

        @tf.function
        def predict_step(inputs_b):
            # sample n times by tiling batch by rows, running
            # predictions for each row
            inputs_tiled = tf.tile(inputs_b, (samples, 1, 1))
            y_pred = tf.sigmoid(self.model(inputs_tiled))[:,Features.FEATURE_IDX.value,:]
            # split up batches into a third dimension and stack them in third dimension
            preds = tf.stack(tf.split(y_pred, samples, axis=0), axis=0)
            return tf.math.reduce_mean(preds, axis=0), tf.math.reduce_std(preds, axis=0)

        for inputs_b, truth_b, weights_b in tqdm.tqdm(iter_.take(batches)):

            # Calculate epistemic uncertainty for batch by iterating over a certain number of times,
            # getting y_preds. You can then calculate the mean and sigma of the predictions,
            # and use this to gather uncertainty: (see http://krasserm.github.io/2019/03/14/bayesian-neural-networks/)
            # inputs, truth, sample_weight
            preds_mean_b, preds_std_b = predict_step(inputs_b)

            preds_mean.append(preds_mean_b)
            preds_std.append(preds_std_b)
            truth.append(truth_b)
            sample_weight.append(weights_b)

        # concat all results
        preds_mean = tf.concat(preds_mean, axis=0)
        preds_std = tf.concat(preds_std, axis=0)

        truth = tf.concat(truth, axis=0)
        sample_weight = tf.concat(sample_weight, axis=0)

        # trim off extra from last batch
        truth = truth[:num_samples, :]
        preds_mean = preds_mean[:num_samples, :]
        preds_std = preds_std[:num_samples, :]
        sample_weight = sample_weight[:num_samples, :]

        # reset truth back to 0 to compute metrics
        # sample weights will rule these out anyways when computing metrics
        truth_reset = np.copy(truth)
        truth_reset[truth_reset < Label.UNBOUND.value] = 0

        # do not continue to calculate metrics. Just return predictions and true values
        if (not calculate_metrics):
            return {
                'preds_mean': preds_mean,
                'preds_std': preds_std,
                'truth': truth,
                'weights': sample_weight,
                'assay_dict': None,
                'auROC': None,
                'auPRC': None
            }

        assert(preds_mean.shape == sample_weight.shape)

        try:

            # try/accept for cases with only one class (throws ValueError)
            assay_dict = get_performance(self.assaymap, preds_mean, truth_reset, sample_weight)

            # calculate averages
            auROC = np.nanmean(list(map(lambda x: x['AUC'],assay_dict.values())))
            auPRC = np.nanmean(list(map(lambda x: x['auPRC'],assay_dict.values())))
            avgGINI = np.nanmean(list(map(lambda x: x['GINI'],assay_dict.values())))

            tf.compat.v1.logging.info("macro auROC:     " + str(auROC))
            tf.compat.v1.logging.info("auPRC:     " + str(auPRC))
            tf.compat.v1.logging.info("GINI:     " + str(avgGINI))
        except ValueError as v:
            auROC = None
            auPRC = None
            tf.compat.v1.logging.info("Failed to calculate metrics")

        return {
            'preds_mean': preds_mean,
            'preds_std': preds_std,
            'truth': truth,
            'weights': sample_weight,
            'assay_dict': assay_dict,
            'auROC': auROC,
            'auPRC': auPRC
        }

    def score_whole_genome(self, chromatin_peak_file,
                       file_prefix,
                       chrs=None,
                       all_data = None):
        """
        Runs a whole genome scan for all available genomic regions in the dataset (about 3.2Million regions)
        Takes about 1 hour.

        Args:
            :param chromatin_peak_file: narrowpeak or bed file containing chromatin accessibility to score
            :param file_prefix: path to save compressed numpy file to. Adds '.npz' extension.
            :param chroms: list of chromosome names to score. If none, scores all chromosomes.
            :param all_data: for testing. If none, generates a concatenated matrix of all data when called.

        """

        # get peak_vector, which is a vector matching train set. Some peaks will not overlap train set,
        # and their indices are stored in missing_idx for future use
        peak_vector_chromatin, _ = bedFile2Vector(chromatin_peak_file, self.regionsFile)

        liRegions = enumerate(load_bed_regions(self.regionsFile))

        # filter liRegions by chrs
        if chrs is not None:
            liRegions = [i for i in liRegions if i[1].chrom in chrs]

        # get indices to score
        idx = np.array([i[0] for i in liRegions])
        liRegions = [i[1] for i in liRegions]

        print("scoring %i regions" % idx.shape[0])

        if all_data is None:
            all_data = concatenate_all_data(self.data, self.regionsFile)

        # tuple of means and stds
        predictions = self.eval_vector(all_data, peak_vector_chromatin, idx)
        print("finished predictions...", predictions[0].shape)

        # zip together means and stdevs for each position in idx

        # return matrix of region, TF information
        npRegions = np.array(list(map(lambda x: np.array([x.chrom, x.start, x.end]),liRegions)))
        # TODO turn into right types (all strings right now)
        means = np.concatenate([npRegions, predictions[0]], axis=1)
        stds = np.concatenate([npRegions, predictions[1]], axis=1)

        # can load back in using:
        # > loaded = np.load('file_prefix.npz')
        # > loaded['means'], loaded['stds']
        np.savez_compressed(file_prefix, means = means, stds=stds,
                            names=np.array(['chr','start','end'] + list(self.assaymap)[1:]))

        print("columns for matrices are chr, start, end, %s" % ", ".join(list(self.assaymap)[1:]))

    def score_peak_file(self,
                        chromatin_peak_file,
                        regions_peak_file,
                        peak_vector_regions = None,
                        all_peaks_regions = None,
                        all_data = None):
        """ Scores cell type specific predictions on a set of genomic loci for
        a cell type with chromatin accessibility peaks.

        Args:
            :param chromatin_peak_file: bed or narrowpeak file of DNase-seq or ATAC-seq peaks
            :param regions_peak_file: bed file of regions to score
            :param all_peaks_regions: pre-calculated peak regions from regions_peak file. If None, calculates.
            :param all_data: concatenated data vector. If None, calculates from self.data.

        Returns:
            finalDF: pandas dataframe of predictions for each region in regions_peak_file.

        """

        # get peak_vector, which is a vector matching train set. Some peaks will not overlap train set,
        # and their indices are stored in missing_idx for future use
        peak_vector_chromatin, _ = bedFile2Vector(chromatin_peak_file, self.regionsFile)

        if peak_vector_regions == None or all_peaks_regions == None:
            peak_vector_regions, all_peaks_regions = bedFile2Vector(regions_peak_file, self.regionsFile)

        # only select peaks to score
        idx = np.where(peak_vector_regions == True)[0]

        print("scoring %i regions" % idx.shape[0])

        if len(idx) == 0:
            raise ValueError("No positive peaks found in %s" % regions_peak_file)

        if all_data is None:
            all_data = concatenate_all_data(self.data, self.regionsFile)

        # tuple of means and stds
        predictions = self.eval_vector(all_data, peak_vector_chromatin, idx)
        # zip together meands and stdevs for each position in idx
        # shape of predictions is (# available predictions, 2 [mean, std], #TFs )
        predictions = np.array(list(zip(predictions[0], predictions[1])))
        print("finished predictions...", predictions.shape)

        # # get number of factors to fill in if values are missing
        num_factors = predictions.shape[-1]

        # map predictions with genomic position
        liRegions = load_bed_regions(self.regionsFile)
        prediction_positions = itemgetter(*idx)(liRegions)
        # list of (region, mean, stdev) for each prediction
        zipped = list(zip(prediction_positions, predictions))

        # # for each all_peaks, if 1, reduce means for all overlapping peaks in positions
        # # else, set to 0s
        def reduceMeans(peak):

            if (peak[1]):
                # parse region

                # filter overlapping predictions for this peak and take mean
                res = np.array(list(map(lambda k: k[1][0], filter(lambda x: Region.overlaps(peak[0], x[0], 1), zipped))))
                mean = np.mean(res, axis = 0)
                # TODO: there are some cases with no peaks
                if res.shape[0] == 0:
                    return(peak[0], np.zeros(num_factors))
                else:
                    return(peak[0], mean)

            else:
                return(peak[0], np.zeros(num_factors))

        # zip together all intervals being evalutated (all_peaks_regions[0]) and whether or
        # not each evaluated region is present (all_peaks_regions[1])
        #
        # for each evaluated region, reduce means
        grouped = list(map(lambda x: np.matrix(reduceMeans(x)[1]), zip(all_peaks_regions[0], all_peaks_regions[1])))
        final = np.concatenate(grouped, axis=0)

        df = pd.DataFrame(final, columns=list(self.assaymap)[1:])


        # load in peaks to get positions and could be called only once
        # TODO why are you reading this in twice?
        df_pos = pd.read_csv(regions_peak_file, sep="\t", header = None)[[0,1,2]]
        final_df = pd.concat([df_pos, df], axis=1)

        return final_df


class VLP(VariationalPeakModel):
    """ Initializes the Variational peak model. Inherits VariationalPeakModel.
    """
    def __init__(self,
             *args,
             **kwargs):

        """ Creates a new model with 4 layers with 100 unites each.
            To resume model training on an old model, call:
            model = VLP(checkpoint=path_to_saved_model)
        """
        self.layers = 4
        self.num_units = [100, 100, 100, 50]
        self.activation = tf.tanh

        if "checkpoint" in kwargs.keys():
            fileObject = open(kwargs["checkpoint"] + "/model_params.pickle" ,'rb')
            metadata = pickle.load(fileObject)
            VariationalPeakModel.__init__(self, **metadata)
            self.model = tf.keras.models.load_model(kwargs["checkpoint"])

        else:
            VariationalPeakModel.__init__(self, *args, **kwargs)

    def create_model(self):
        """ Create a sequential keras model with dense flipout layers.
        """

        # create a linear stack of layers
        model = tf.keras.Sequential()

        if not isinstance(self.num_units, collections.Iterable):
            self.num_units = [self.num_units] * self.layers
        # Add densely-connected layer class with Flipout estimator to model
        for i in range(self.layers):
            model.add(tfp.layers.DenseFlipout(self.num_units[i], activation = self.activation))

        # output layer
        model.add(tfp.layers.DenseFlipout(self.num_outputs,
                                          activity_regularizer=tf.keras.regularizers.l1_l2(self.l1, self.l2)))

        return model



import os
import sys
import tempfile
import unittest
from epitome.models import *
from epitome import GET_DATA_PATH

class EpitomeTestCase(unittest.TestCase):

	def makeSmallModel(self):

		data_path = GET_DATA_PATH()
		x = os.path.join(data_path, 'valid.npz')
		# load in small validation matrix for test
		sparse_matrix = scipy.sparse.load_npz(x).toarray()
		data = {Dataset.TRAIN: sparse_matrix, Dataset.VALID: sparse_matrix, Dataset.TEST: sparse_matrix}

		
		eligible_cells = ['K562','HepG2','H1','A549','HeLa-S3']
		eligible_assays = ['DNase','CTCF']
		matrix, cellmap, assaymap = get_assays_from_feature_file(os.path.join(data_path, "feature_name"),
                                                         eligible_assays = eligible_assays,
                                  eligible_cells = eligible_cells, min_cells_per_assay = 3, min_assays_per_cell = 1)

		return VLP(data_path,
            		['K562'],
            		matrix,
            		assaymap,
            		cellmap,
            		shuffle_size=2, 
            		batch_size=64, data = data)



r"""
=======================
Vizualization functions
=======================
.. currentmodule:: epitome.viz

.. autosummary::
  :toctree: _generate/

  plot_assay_heatmap
  joint_plot
"""

#####################################################################
################### Visualization functions #########################
#####################################################################

import seaborn as sns; sns.set()
import matplotlib.pyplot as plt
import pandas as pd
import numpy as np
from matplotlib.backends import backend_agg
from matplotlib import figure
import tensorflow as tf

def joint_plot(dict_model1,
               dict_model2,
               metric = "AUC",
               model1_name = "model1",
               model2_name = "model2",
               outlier_filter = None
              ):
    """
    Returns seaborn joint plot of two models.

    Args:
        :param dict_model1: dictionary of TF: metrics for first model. Output from run_predictions().
        :param dict_model2: dictionary of TF: metrics for second model. Output from run_predictions().
        :param metric: metric in dicts. Should be auPRC, AUC, or GINI.
        :param model1_name: string for model 1 name, shown on x axis.
        :param model2_name: string for model 2 name, shown on y axis.
        :param outlier_filter: string filter to label. Defaults to no labels.

    Returns:
        matplotlib axis
    """
    sns.set(style="whitegrid", color_codes=True)

    df1 = pd.DataFrame.from_dict(dict_model1).T
    df2 = pd.DataFrame.from_dict(dict_model2).T

    dataframe_joined = pd.concat([df1, df2], axis=1, sort=False)
    # select metric
    dataframe = dataframe_joined[metric]
    dataframe.columns=[model1_name, model2_name]
    dataframe = dataframe.dropna()

    ax = sns.jointplot(x=list(dataframe[model1_name]), y=list(dataframe[model2_name]), kind="reg", stat_func=None)

    ax.ax_joint.cla()
    ax.set_axis_labels(model1_name, model2_name)
    ax.fig.suptitle(metric)

    def ann(row):
        ind = row[0]
        r = row[1]
        plt.gca().annotate(ind, xy=(r[model1_name], r[model2_name]),
                xytext=(2,2) , textcoords ="offset points", )

    # filter for labels
    if (outlier_filter is not None):
        label_df = dataframe.query(outlier_filter)

        for index, row in label_df.iterrows():
            if (not np.isnan(row[model1_name]) and not np.isnan(row[model2_name])):
                ann((index, row))

    for i,row in dataframe.iterrows():
        color = "blue" if row[model1_name] <  row[model2_name] else "red"
        ax.ax_joint.plot(row[model1_name], row[model2_name], color=color, marker='o')

    return ax



def plot_assay_heatmap(matrix, cellmap, assaymap):
    """ Plots a matrix of available assays from available cells.

    Args:
        :param matrix: numpy matrix of indices that index into Epitome data
        :param cellmap: map of cells indexing into rows of matrix
        :param assaymap: map of assays indexing into columns of matrix
    """

    nv_assaymap = {v: k for k, v in assaymap.items()}

    fig = plt.figure(figsize = (20,10))
    ax = fig.add_subplot(1,1,1)
    ax.set_aspect('equal')
    plt.xticks(np.arange(len(assaymap)), rotation = 90)
    ax.set_xticklabels(assaymap.keys())
    plt.yticks(np.arange(len(cellmap)))
    ax.set_yticklabels(cellmap.keys())

    plt.imshow(matrix!=-1)




#####################################################################
############## Network visualization helper functions ###############
#####################################################################

def plot_uncertainty(preds_mean, preds_std, truth, title):
    """
    Plot means vs stds for regions

    Args:
    :param preds_mean: prediction means
    :param preds_std: prediction standard devations
    :param truth: 0/1 real values
    :param title: plot title

    """
    xdata = np.arange(0, preds_mean.shape[0])
    # Visualize the result
    plt.plot(xdata, truth, 'ro')
    plt.plot(xdata, preds_mean, '-', color='gray')

    plt.xlabel("genomic region")
    plt.ylabel("predictions (dark grey), stdev (light grey)")

    plt.fill_between(xdata, preds_mean - preds_std, preds_mean + preds_std,
                     color='gray', alpha=0.2)

    plt.title(title)
    plt.show()


def number_to_bp(n):
    """ converts bp number to short string
    :param n: number in base pairs
    :return string of number with kbp or Mbp suffix
    """
    n = str(n)

    if len(n) < 4:
        return n
    elif len(n) == 4:
        return "%skbp" % n[0]
    elif len(n) == 5:
        return "%skbp" % n[0:1]
    elif len(n) == 6:
        return "%skbp" % n[0:2]
    elif len(n) == 7:
        return "%sMbp" % n[0]
    else:
        raise



def heatmap_aggreement_from_model_weights(model):
    """
    Plots seaborn heatmap for DNase weights of first layer in network.
    Plots one heatmap for each celltype used in the features for training (about 10-13).

    :param model: an Epitome model
    """

    # get weights
    with model.graph.as_default():
        with tf.variable_scope('layer0', reuse=True):
            w = tf.get_variable('kernel')
            weights = model.sess.run(w)

    dnases = weights[len(model.assaymap) * (len(model.cellmap) - 1 - len(model.test_celltypes)): , :]

    num_radii = 2*len(model.radii)

    xtick_labels = [x * 200 for x in model.radii]
    xtick_labels = [number_to_bp(x) for x in xtick_labels]

    for i in range((len(model.cellmap) - 1 - len(model.test_celltypes))):
        # dnases are just at the end
        dnases = x[len(model.assaymap) * (len(model.cellmap) - 2): , :]

        heatmap = dnases[i*num_radii + int(num_radii/2):i*num_radii + num_radii, :].T
        ax = sns.heatmap(heatmap, cmap='bwr_r', vmin=np.min(dnases), vmax=np.max(dnases))
        ax.set_xlabel("Distance from peak (bp)")
        ax.set_ylabel("Unit")
        ax.set_xticklabels(xtick_labels, rotation=-60)
        plt.show()



import matplotlib.pyplot as plt
import matplotlib.lines as mlines
import matplotlib.transforms as mtransforms

def calibration_plot(truth, preds, assay_dict, list_assaymap):
    """
    Creates an xy scatter plot for predicted probability vs true probability.
    Adds a separate set of points for each transcription factor.

    Args:
        :param truth: matrix of n samples by t TFs
        :param preds: matrix same size as truth
        :param assay_dict: dictionary of TFs and scores
        :param: list_assaymap: list of assay names for data matrix
    """

    fig, ax = plt.subplots()
    # only these two lines are calibration curves
    for i in range(truth.shape[1]):
        logreg_y, logreg_x = calibration_curve(truth[:,i], preds[:,i], n_bins=10)

        if (not np.isnan(assay_dict[list_assaymap[i+1]]["AUC"])): # test cell type does not have all factors!
            plt.plot(logreg_x,logreg_y, marker='o', linewidth=1, label=list_assaymap[i+1])


    # reference line, legends, and axis labels
    line = mlines.Line2D([0, 1], [0, 1], color='black')
    transform = ax.transAxes
    line.set_transform(transform)
    ax.add_line(line)
    fig.suptitle('Calibration plot for test regions')
    ax.set_xlabel('Predicted probability')
    ax.set_ylabel('True probability in each bin')

    ax.legend(loc='center left', bbox_to_anchor=(1, 0.5),
              ncol=2, fancybox=True, shadow=True)

    plt.show()




############################################################################
###################### Visualizing model internals #########################
############################################################################

def plot_weight_posteriors(names, qm_vals, qs_vals, fname=None):
    """Save a PNG plot with histograms of weight means and stddevs.
    From https://github.com/tensorflow/probability/blob/master/tensorflow_probability/examples/bayesian_neural_network.py.
    Requires that model has been trained for at least 1 iteration to correctly instantiate kernel posterior.
    To collect parameters, run:
    
    ```python
    
    model = VLP(...)
    names, means, stds = model.get_weight_parameters()
    
    ```
    
    Args:
    names: A Python `iterable` of `str` variable names.
    qm_vals: A Python `iterable`, the same length as `names`,
      whose elements are Numpy `array`s, of any shape, containing
      posterior means of weight varibles.
    qs_vals: A Python `iterable`, the same length as `names`,
      whose elements are Numpy `array`s, of any shape, containing
      posterior standard deviations of weight varibles.
    fname: Python `str` filename to save the plot to.
    """
    fig = plt.figure(figsize=(6, 3))

    ax = fig.add_subplot(1, 2, 1)
    for n, qm in zip(names, qm_vals):
        sns.distplot(tf.reshape(qm, [-1]), ax=ax, label=n)
    ax.set_title("weight means")
    ax.set_xlim([-1.5, 1.5])
    ax.legend()

    ax = fig.add_subplot(1, 2, 2)
    for n, qs in zip(names, qs_vals):
        sns.distplot(tf.reshape(qs, [-1]), ax=ax)
    ax.set_title("weight stddevs")
    ax.set_xlim([0, 0.2])

    fig.tight_layout()
    if fname != None:
        fig.savefig(fname)
        print("saved {}".format(fname))

    return fig



"""
Functions for data generators.
"""

import numpy as np
import tensorflow as tf
from .constants import *
from .functions import *
import epitome.iio as iio
import glob

######################### Original Data Generator: Only peak based #####################
np.random.seed(0) # to keep np.random.choice consistent

def load_data(data, 
                 label_cell_types,  # used for labels. Should be all for train/eval and subset for test
                 eval_cell_types,   # used for rotating features. Should be all - test for train/eval
                 matrix,
                 assaymap,
                 cellmap,
                 radii,
                 **kwargs):
    
    # AM 5/20/2019. This is enforcing exclusive DNase bins and will make 
    # interpretation of DNase weights easier. It does not add performance benefit
    # over inclusive bins, which was used in the original model.
    exclusive = True

    """
    Takes Deepsea data and calculates distance metrics from cell types whose locations
    are specified by label_cell_indices, and the other cell types in the set. Label space is only one cell type.
     TODO AM 3/7/2019
    :param data: dictionary of matrices. Should have keys x and y. x contains n by 1000 rows. y contains n by 919 labels.
    :param label_cell_types: list of cell types to be rotated through and used as labels (subset of eval_cell_types)
    :param eval_cell_types: list of cell types to be used in evaluation (includes label_cell_types)
    :param matrix: matrix of celltype, assay positions
    :param assaymap: map of column assay positions in matrix
    :param cellmap: map of row cell type positions in matrix
    :param radii: radii to compute dnase distances from
    :param kwargs: kargs

    :returns: generator of data with three elements:
        1. record features
        2. record labels for a given cell type
        3. 0/1 mask of labels that have validation data. For example, if this record is for celltype A549,
        and A549 does not have data for ATF3, there will be a 0 in the position corresponding to the label space.
    """

    # Running in TRAIN, VALID, or TEST?    
    mode = kwargs.get("mode")
    # specifies the indices to generate records.
    # can be used for debug purposes, or to evaluate
    # only specific regions in a vector
    # TODO AM 4/17/2019: move this to an actual parameter
    indices = kwargs.get("indices")
    
    if (not isinstance(indices, np.ndarray) and not isinstance(indices, list)):
        # model performs better when there are less 0s
        if mode == Dataset.TRAIN:
            feature_indices = np.concatenate(list(map(lambda c: get_y_indices_for_cell(matrix, cellmap, c), 
                                     list(cellmap))))
            feature_indices = feature_indices[feature_indices != -1]

            # need to re-proportion the indices to equalize positives
            if (len(list(assaymap)) > 2):

                # get sums for each feature in the dataset
                rowsums = np.sum(data[feature_indices,:], axis=1) 

                # multiply data by row scaling factor
                scale_factor = 1/rowsums
                scaled = data[feature_indices,:] * scale_factor[:, np.newaxis] 

                # indices where sum > 0
                indices_zero = np.where(np.sum(scaled, axis=0) > 0)[0]
                # then filter indices by probabilities inversely proportional to frequency
                indices = np.random.choice(indices_zero, int(indices_zero.shape[0] * 0.4), p=(np.sum(scaled, axis=0)/np.sum(scaled))[indices_zero])

            else:
                # single TF model
                feature_indices = np.concatenate(list(map(lambda c: get_y_indices_for_cell(matrix, cellmap, c), 
                                                     list(cellmap))))

                # chop of DNase
                TF_indices = feature_indices.reshape([len(cellmap),len(assaymap)])[:,1]
                
                TF_indices =  TF_indices[TF_indices != -1]
                feature_indices = feature_indices[feature_indices != -1]

                # sites where TF is in at least 1 cell line
                positive_indices = np.where(np.sum(data[TF_indices,:], axis=0) > 1)[0]

                indices_probs = np.ones([data.shape[1]])
                indices_probs[positive_indices] = 0
                indices_probs = indices_probs/np.sum(indices_probs, keepdims=1)

                # randomly select 10 fold sites where TF is not in any cell line
                negative_indices = np.random.choice(np.arange(0,data.shape[1]), positive_indices.shape[0] * 10,p=indices_probs)
                indices = np.sort(np.concatenate([negative_indices, positive_indices]))

                
        else:
            indices = range(0, data.shape[-1]) # if not defined, set to all points
    
    if (not isinstance(mode, Dataset)):
        raise ValueError("mode is not a Dataset enum")
        
    if (mode == Dataset.RUNTIME):
        label_cell_types = ["PLACEHOLDER_CELL"]
        dnase_vector = kwargs.get("dnase_vector")
        random_cell = list(cellmap)[0] # placeholder to get label vector length
        
    print("using %s as labels for mode %s" % (label_cell_types, mode))
    
    # string of radii for meta data labeling
    radii_str = list(map(lambda x: "DNASE_RADII_%i" % x, radii))
        
    if (mode == Dataset.TEST or mode == Dataset.RUNTIME):
        # Drop cell types with the least information (TODO AM 4/1/2019 this could do something smarter)
        
        # make dictionary of eval_cell_type: assay count and sort in decreasing order
        tmp = matrix.copy()
        tmp[tmp>= 0] = 1
        tmp[tmp== -1] = 0
        sums = np.sum(tmp, axis = 1)
        cell_assay_counts = zip(list(cellmap), sums)
        cell_assay_counts = sorted(cell_assay_counts, key = lambda x: x[1])
        # filter by eval_cell_types
        cell_assay_counts = list(filter(lambda x: x[0] in eval_cell_types, cell_assay_counts))
        
        # remove cell types with smallest number of factors
        eval_cell_types = list(eval_cell_types)
        [eval_cell_types.remove(i[0]) for i in cell_assay_counts[0:len(label_cell_types)]]
        del tmp
        del cell_assay_counts
        
    def g():
        for i in indices: # for all records specified
            for (cell) in label_cell_types: # for all cell types to be used in labels
                dnases = [] 
                dnases_double_positive = []
                dnases_agreement = []
                
                # cells to be featurized
                feature_cells = list(eval_cell_types)
                
                # try to remove cell if it is in the possible list of feature cell types
                try:
                    feature_cells.remove(cell) 
                except ValueError:
                    pass  # do nothing!
                                
                # features from all remaining cells not in label set
                feature_cell_indices_list = list(map(lambda c: get_y_indices_for_cell(matrix, cellmap, c), 
                                                     feature_cells))
                feature_cell_indices = np.array(feature_cell_indices_list).flatten()
                
                # labels for this cell
                if (mode != Dataset.RUNTIME):
                    label_cell_indices = get_y_indices_for_cell(matrix, cellmap, cell)
                    label_cell_indices_no_dnase = np.delete(label_cell_indices, [0])

                    # Copy assay_index_no_dnase and turn into mask of 0/1 for whether data for this cell type for
                    # a given label is available.
                    assay_mask = np.copy(label_cell_indices_no_dnase)
                    assay_mask[assay_mask == -1] = 0
                    assay_mask[assay_mask > 0] = 1
                    
                else:
                    label_count = len(get_y_indices_for_cell(matrix, cellmap, random_cell))-1
                    
                    # Mask and labels are all 0's because labels are missing during runtime
                    garbage_labels = assay_mask = np.zeros(label_count)

                # get dnase indices for cell types that are going to be features
                dnase_indices = np.array([x[0] for x in feature_cell_indices_list])
                
                for r, radius in enumerate(radii):
                    
                    min_radius = max(0, i - radius + 1)
                    max_radius = min(i+radius, data.shape[1])
                    
                    # if exclusive == True, then do not featurize chromatin regions
                    # that were considered in smaller radii
                    if (exclusive and r != 0):
                        radius_range_1 = np.arange(min_radius, max(0, i - radii[r-1]+1))
                        radius_range_2 = np.arange(i+radii[r-1], max_radius)
                        
                        radius_range = np.concatenate([radius_range_1, radius_range_2])
                    else:
                        
                        radius_range = np.arange(min_radius, max_radius)
                        
                        
                    ####################################################################
                    
                    # use DNase vector, if it is provided
                    if (mode == Dataset.RUNTIME):

                        # within the radius, fraction of places where they are both 1
                        dnase_double_positive = np.average(data[dnase_indices[:,None],radius_range]*
                                                 dnase_vector[radius_range], axis=1)

                        # within the radius, fraction of places where they are both equal (0 or 1)
                        dnase_agreement = np.average(data[dnase_indices[:,None],radius_range]==
                                                 dnase_vector[radius_range], axis=1)

                    else:
                        # within the radius, fraction of places where they are both 1
                        # label_cell_index[0] == DNase location for specific cell type
                        dnase_double_positive = np.average(data[dnase_indices[:,None],radius_range]*
                                                 data[label_cell_indices[0],radius_range], axis=1)

                        # within the radius, fraction of places where they are both equal (0 or 1)
                        dnase_agreement = np.average(data[dnase_indices[:,None],radius_range]==
                                                 data[label_cell_indices[0],radius_range], axis=1)
                        
                    dnases_double_positive.extend(dnase_double_positive)
                    dnases_agreement.extend(dnase_agreement)
                        
                # rehape agreement DNase to Radii by feature_cells
                dnases_agreement_reshaped = np.array(dnases_agreement).reshape([len(radii), len(feature_cells)])
                dnases_double_positive_reshaped = np.array(dnases_double_positive).reshape([len(radii), len(feature_cells)])
                dnase_means = np.mean(dnases_agreement_reshaped, axis = 0)

                ######### reorder cells by similarity ################
                ## This was added 5/30/2019. It seems to *maybe help 
                ## a little bit on cell types not seen in the model.
                ## This makes sense because cell types are now ordered
                ## by similarity and keep some spacial positioning 
                ## based on the similarity to the new cell. 
                best_indices = (-dnase_means).argsort()

                dnases.extend(dnases_double_positive_reshaped[:,best_indices].flatten())
                dnases.extend(dnases_agreement_reshaped[:,best_indices].flatten())

                feature_cell_indices = feature_cell_indices.reshape([len(feature_cells), len(assaymap)])[best_indices,:].flatten()
                ######## End reorder #################################                                                   
                                                                        
                                                                        
                # Extract features
                features = data[feature_cell_indices,i]
                
                # concatenate features and DNases
                x_data = np.concatenate([features, dnases])
                
                # mask for x_data. 0 = do not mask, 1 = mask.
                x_mask = np.zeros(x_data.shape[0])
                x_mask[np.where(feature_cell_indices == -1)[0]] = True # assign mask to missing features

                # There can be NaNs in the DNases for edge cases (when the radii extends past the end of the chr).
                # Mask these values in the first row of tmp
                x_mask[np.where(np.isnan(x_data))[0]] = True # assign mask to missing DNase values
                x_data[np.where(x_mask == True)[0]] = 0 # set all UNKs to 0
                
                # mask x data by which factors are available for a cell type
                x_data_masked = np.vstack([x_mask, x_data]) # top row 0 = mask, bottom row 1 = data
                        
                if (mode != Dataset.RUNTIME):
                    labels = data[label_cell_indices_no_dnase,i]

                else: # used when just predicting
                    # The features going into the example.
                    labels = garbage_labels # all 0's
                    
                yield (x_data_masked, labels, assay_mask)

    return g



def load_v_data(data, 
                 label_cell_types,  # used for labels. Should be all for train/eval and subset for test
                 eval_cell_types,   # used for rotating features. Should be all - test for train/eval
                 matrix,
                 assaymap,
                 cellmap,
                 radii,
                 **kwargs):
    
    # AM 5/20/2019. This is enforcing exclusive DNase bins and will make 
    # interpretation of DNase weights easier. It does not add performance benefit
    # over inclusive bins, which was used in the original model.
    exclusive = True

    """
    Takes Deepsea data and calculates distance metrics from cell types whose locations
    are specified by label_cell_indices, and the other cell types in the set. Label space is only one cell type.
     TODO AM 3/7/2019
    :param data: dictionary of matrices. Should have keys x and y. x contains n by 1000 rows. y contains n by 919 labels.
    :param label_cell_types: list of cell types to be rotated through and used as labels (subset of eval_cell_types)
    :param eval_cell_types: list of cell types to be used in evaluation (includes label_cell_types)
    :param matrix: matrix of celltype, assay positions
    :param assaymap: map of column assay positions in matrix
    :param cellmap: map of row cell type positions in matrix
    :param radii: radii to compute dnase distances from
    :param kwargs: kargs

    :returns: generator of data with three elements:
        1. record features
        2. record labels for a given cell type
        3. 0/1 mask of labels that have validation data. For example, if this record is for celltype A549,
        and A549 does not have data for ATF3, there will be a 0 in the position corresponding to the label space.
    """
    # set all UNKs features to TF mean
    def TF_mean(i, idx, cell_indices):
        """
        Gets the mean TF binding score for an element in this location
        """
        # compute means at row i in data for all assays
        indices = matrix[cell_indices,:].T.flatten()
        tmp1 = data[:,i][indices].astype(None)
        tmp1[tmp1 == Label.UNK.value] = np.nan
        tmp1[tmp1 == Label.IMPUTED_BOUND.value] = Label.BOUND.value
        tmp1[tmp1 == Label.IMPUTED_UNBOUND.value] = Label.UNBOUND.value
        means = np.nanmean(tmp1.reshape([len(cell_indices),  matrix.shape[1]]), axis=0) 
        
        TF_idx = idx % matrix.shape[1]
        return means[TF_idx]

    # Running in TRAIN, VALID, or TEST?    
    mode = kwargs.get("mode")
    # specifies the indices to generate records.
    # can be used for debug purposes, or to evaluate
    # only specific regions in a vector
    # TODO AM 4/17/2019: move this to an actual parameter
    indices = kwargs.get("indices")
    
    if (not isinstance(indices, np.ndarray) and not isinstance(indices, list)):
        indices = range(0, data.shape[-1]) # if not defined, set to all points
    
    if (not isinstance(mode, Dataset)):
        raise ValueError("mode is not a Dataset enum")
        
    if (mode == Dataset.RUNTIME):
        label_cell_types = ["PLACEHOLDER_CELL"]
        dnase_vector = kwargs.get("dnase_vector")
        random_cell = list(cellmap)[0] # placeholder to get label vector length
        
    print("using %s as labels for mode %s" % (label_cell_types, mode))
    
    # string of radii for meta data labeling
    radii_str = list(map(lambda x: "DNASE_RADII_%i" % x, radii))
        
    if (mode == Dataset.TEST or mode == Dataset.RUNTIME):
        # Drop cell types with the least information (TODO AM 4/1/2019 this could do something smarter)
        
        # make dictionary of eval_cell_type: assay count and sort in decreasing order
        tmp = matrix.copy()
        tmp[tmp >= Label.UNBOUND.value] = 1
        tmp[tmp == Label.UNK.value] = 0 
        sums = np.sum(tmp, axis = 1)
        cell_assay_counts = zip(list(cellmap), sums)
        cell_assay_counts = sorted(cell_assay_counts, key = lambda x: x[1])
        # filter by eval_cell_types
        cell_assay_counts = list(filter(lambda x: x[0] in eval_cell_types, cell_assay_counts))
        
        # remove cell types with smallest number of factors
        eval_cell_types = eval_cell_types.copy()
        [eval_cell_types.remove(i[0]) for i in cell_assay_counts[0:len(label_cell_types)]]
        del tmp
        del cell_assay_counts
        
        
    def g():
        for i in indices: # for all records specified
            for (cell) in label_cell_types: # for all cell types to be used in labels
                dnases = [] 
                dnases_double_positive = []
                dnases_agreement = []
                
                # cells to be featurized
                feature_cells = eval_cell_types.copy()
                
                # try to remove cell if it is in the possible list of feature cell types
                try:
                    feature_cells.remove(cell) 
                except ValueError:
                    pass  # do nothing!
                                
                # features from all remaining cells not in label set
                feature_cell_indices_list = list(map(lambda c: get_y_indices_for_cell(matrix, cellmap, c), 
                                                     feature_cells))
                feature_cell_indices = np.array(feature_cell_indices_list).flatten()
                
                                
                # labels for this cell
                if (mode != Dataset.RUNTIME):
                    label_cell_indices = get_y_indices_for_cell(matrix, cellmap, cell)
                    label_cell_indices_no_dnase = np.delete(label_cell_indices, [0])
                    labels = data[label_cell_indices_no_dnase,i]

                    # Copy assay_index_no_dnase and turn into mask of 0/1 for whether data for this cell type for
                    # a given label is available.
                    assay_mask = np.ones(labels.shape)
                    # set mask over missing AND imputed mask
                    assay_mask[labels < Label.UNBOUND.value] = 0
                else:
                    label_count = len(get_y_indices_for_cell(matrix, cellmap, random_cell))-1
                    
                    # Mask and labels are all 0's because labels are missing during runtime
                    garbage_labels = assay_mask = np.zeros(label_count)
                    # The features going into the example.
                    labels = garbage_labels # all 0's

                # get dnase indices for cell types that are going to be features
                dnase_indices = np.array([x[0] for x in feature_cell_indices_list])
                
                for r, radius in enumerate(radii):
                    
                    min_radius = max(0, i - radius + 1)
                    max_radius = min(i+radius, data.shape[1])
                    
                    # if exclusive == True, then do not featurize chromatin regions
                    # that were considered in smaller radii
                    if (exclusive and r != 0):
                        radius_range_1 = np.arange(min_radius, max(0, i - radii[r-1]+1))
                        radius_range_2 = np.arange(i+radii[r-1], max_radius)
                        
                        radius_range = np.concatenate([radius_range_1, radius_range_2])
                    else:
                        
                        radius_range = np.arange(min_radius, max_radius)
                        
                        
                    ####################################################################
                    
                    # use DNase vector, if it is provided
                    if (mode == Dataset.RUNTIME):

                        # within the radius, fraction of places where they are both 1
                        dnase_double_positive = np.average(data[dnase_indices[:,None],radius_range]*
                                                 dnase_vector[radius_range], axis=1)

                        # within the radius, fraction of places where they are both equal (0 or 1)
                        dnase_agreement = np.average(data[dnase_indices[:,None],radius_range]==
                                                 dnase_vector[radius_range], axis=1)

                    else:
                        # within the radius, fraction of places where they are both 1
                        # label_cell_index[0] == DNase location for specific cell type
                        dnase_double_positive = np.average(data[dnase_indices[:,None],radius_range]*
                                                 data[label_cell_indices[0],radius_range], axis=1)

                        # within the radius, fraction of places where they are both equal (0 or 1)
                        dnase_agreement = np.average(data[dnase_indices[:,None],radius_range]==
                                                 data[label_cell_indices[0],radius_range], axis=1)
                        
                    dnases_double_positive.extend(dnase_double_positive)
                    dnases_agreement.extend(dnase_agreement)
                        
                # rehape agreement DNase to Radii by feature_cells
                dnases_agreement_reshaped = np.array(dnases_agreement).reshape([len(radii), len(feature_cells)])
                dnases_double_positive_reshaped = np.array(dnases_double_positive).reshape([len(radii), len(feature_cells)])
                dnase_means = np.mean(dnases_agreement_reshaped, axis = 0)

                ######### reorder cells by similarity ################
                ## This was added 5/30/2019. It seems to *maybe help 
                ## a little bit on cell types not seen in the model.
                ## This makes sense because cell types are now ordered
                ## by similarity and keep some spacial positioning 
                ## based on the similarity to the new cell. 
                best_indices = (-dnase_means).argsort()

                dnases.extend(dnases_double_positive_reshaped[:,best_indices].flatten())
                dnases.extend(dnases_agreement_reshaped[:,best_indices].flatten())

                feature_cell_indices = feature_cell_indices.reshape([len(feature_cells), len(assaymap)])[best_indices,:].flatten()
                ######## End reorder #################################                                                   
                                                                        
                                                                        
                # Extract features
                features = data[feature_cell_indices,i]
                
                # concatenate features and DNases
                x_data = np.concatenate([features, dnases])
                
                # mask for x_data. 0 = do not mask, 1 = mask.
                x_mask = np.zeros(x_data.shape[0])
                # mask imputed (bound and unbound) and missing features
                x_mask[np.where(features == Label.UNK.value)[0]] = True # assign mask to missing features

                # There can be NaNs in the DNases for edge cases (when the radii extends past the end of the chr).
                # Mask these values in the first row of tmp
                x_mask[np.where(np.isnan(x_data))[0]] = True # assign mask to missing DNase values
                x_data[np.where(x_mask == True)[0]] = 0 # set all UNKs to 0
                
                # reset imputed values to normal known values in x_data
                x_data[np.where(x_data == Label.IMPUTED_BOUND.value)[0]] = Label.BOUND.value 
                x_data[np.where(x_data == Label.IMPUTED_UNBOUND.value)[0]] = Label.UNBOUND.value 
                
                # recalculate missing values as means
                masked_indices = np.where(x_mask == True)[0][:len(feature_cell_indices)]
                cellmap_idx = list(map(lambda x: cellmap[x], feature_cells))
                x_data[masked_indices] = TF_mean(i, masked_indices, cellmap_idx)
                
                yield (x_data, labels, assay_mask)

    return g

def shuffled_data(data, 
                 label_cell_types,  # used for labels. Should be all for train/eval and subset for test
                 eval_cell_types,   # used for rotating features. Should be all - test for train/eval
                 matrix,
                 assaymap,
                 cellmap,
                 radii,
                 **kwargs):
    
    # AM 5/20/2019. This is enforcing exclusive DNase bins and will make 
    # interpretation of DNase weights easier. It does not add performance benefit
    # over inclusive bins, which was used in the original model.
    exclusive = True

    """
    Shuffles features. Used to test uncertainty in model.
    :param data: dictionary of matrices. Should have keys x and y. x contains n by 1000 rows. y contains n by 919 labels.
    :param label_cell_types: list of cell types to be rotated through and used as labels (subset of eval_cell_types)
    :param eval_cell_types: list of cell types to be used in evaluation (includes label_cell_types)
    :param matrix: matrix of celltype, assay positions
    :param assaymap: map of column assay positions in matrix
    :param cellmap: map of row cell type positions in matrix
    :param radii: radii to compute dnase distances from
    :param kwargs: kargs

    :returns: generator of data with three elements:
        1. record features
        2. record labels for a given cell type
        3. 0/1 mask of labels that have validation data. For example, if this record is for celltype A549,
        and A549 does not have data for ATF3, there will be a 0 in the position corresponding to the label space.
    """

    # Running in TRAIN, VALID, or TEST?    
    mode = kwargs.get("mode")
    # specifies the indices to generate records.
    # can be used for debug purposes, or to evaluate
    # only specific regions in a vector
    # TODO AM 4/17/2019: move this to an actual parameter
    indices = kwargs.get("indices")
    
    if (not isinstance(indices, np.ndarray) and not isinstance(indices, list)):
        indices = range(0, data.shape[-1]) # if not defined, set to all points
    
    if (not isinstance(mode, Dataset)):
        raise ValueError("mode is not a Dataset enum")
        
    if (mode == Dataset.RUNTIME):
        label_cell_types = ["PLACEHOLDER_CELL"]
        dnase_vector = kwargs.get("dnase_vector")
        random_cell = list(cellmap)[0] # placeholder to get label vector length
        
    print("using %s as labels for mode %s" % (label_cell_types, mode))
    
    # string of radii for meta data labeling
    radii_str = list(map(lambda x: "DNASE_RADII_%i" % x, radii))
        
    if (mode == Dataset.TEST or mode == Dataset.RUNTIME):
        # Drop cell types with the least information (TODO AM 4/1/2019 this could do something smarter)
        
        # make dictionary of eval_cell_type: assay count and sort in decreasing order
        tmp = matrix.copy()
        tmp[tmp>= 0] = 1
        tmp[tmp== -1] = 0
        sums = np.sum(tmp, axis = 1)
        cell_assay_counts = zip(list(cellmap), sums)
        cell_assay_counts = sorted(cell_assay_counts, key = lambda x: x[1])
        # filter by eval_cell_types
        cell_assay_counts = list(filter(lambda x: x[0] in eval_cell_types, cell_assay_counts))
        
        # remove cell types with smallest number of factors
        eval_cell_types = list(eval_cell_types)
        [eval_cell_types.remove(i[0]) for i in cell_assay_counts[0:len(label_cell_types)]]
        del tmp
        del cell_assay_counts
        
        
    def g():
        for i in indices: # for all records specified
            for (cell) in label_cell_types: # for all cell types to be used in labels
                dnases = [] 
                dnases_double_positive = []
                dnases_agreement = []
                
                # cells to be featurized
                feature_cells = list(eval_cell_types)
                
                # try to remove cell if it is in the possible list of feature cell types
                try:
                    feature_cells.remove(cell) 
                except ValueError:
                    pass  # do nothing!
                                
                # features from all remaining cells not in label set
                feature_cell_indices_list = list(map(lambda c: get_y_indices_for_cell(matrix, cellmap, c), 
                                                     feature_cells))
                feature_cell_indices = np.array(feature_cell_indices_list).flatten()
                
                # labels for this cell
                if (mode != Dataset.RUNTIME):
                    label_cell_indices = get_y_indices_for_cell(matrix, cellmap, cell)
                    label_cell_indices_no_dnase = np.delete(label_cell_indices, [0])

                    # Copy assay_index_no_dnase and turn into mask of 0/1 for whether data for this cell type for
                    # a given label is available.
                    assay_mask = np.copy(label_cell_indices_no_dnase)
                    assay_mask[assay_mask == -1] = 0
                    assay_mask[assay_mask > 0] = 1
                    
                else:
                    label_count = len(get_y_indices_for_cell(matrix, cellmap, random_cell))-1
                    
                    # Mask and labels are all 0's because labels are missing during runtime
                    garbage_labels = assay_mask = np.zeros(label_count)

                # get dnase indices for cell types that are going to be features
                dnase_indices = np.array([x[0] for x in feature_cell_indices_list])
                
                for r, radius in enumerate(radii):
                    
                    min_radius = max(0, i - radius + 1)
                    max_radius = min(i+radius, data.shape[1])
                    
                    # if exclusive == True, then do not featurize chromatin regions
                    # that were considered in smaller radii
                    if (exclusive and r != 0):
                        radius_range_1 = np.arange(min_radius, max(0, i - radii[r-1]+1))
                        radius_range_2 = np.arange(i+radii[r-1], max_radius)
                        
                        radius_range = np.concatenate([radius_range_1, radius_range_2])
                    else:
                        
                        radius_range = np.arange(min_radius, max_radius)
                        
                        
                    ####################################################################
                    
                    # use DNase vector, if it is provided
                    if (mode == Dataset.RUNTIME):

                        # within the radius, fraction of places where they are both 1
                        dnase_double_positive = np.average(data[dnase_indices[:,None],radius_range]*
                                                 dnase_vector[radius_range], axis=1)

                        # within the radius, fraction of places where they are both equal (0 or 1)
                        dnase_agreement = np.average(data[dnase_indices[:,None],radius_range]==
                                                 dnase_vector[radius_range], axis=1)

                    else:
                        # within the radius, fraction of places where they are both 1
                        # label_cell_index[0] == DNase location for specific cell type
                        dnase_double_positive = np.average(data[dnase_indices[:,None],radius_range]*
                                                 data[label_cell_indices[0],radius_range], axis=1)

                        # within the radius, fraction of places where they are both equal (0 or 1)
                        dnase_agreement = np.average(data[dnase_indices[:,None],radius_range]==
                                                 data[label_cell_indices[0],radius_range], axis=1)
                        
                    dnases_double_positive.extend(dnase_double_positive)
                    dnases_agreement.extend(dnase_agreement)
                        
                # rehape agreement DNase to Radii by feature_cells
                dnases_agreement_reshaped = np.array(dnases_agreement).reshape([len(radii), len(feature_cells)])
                dnases_double_positive_reshaped = np.array(dnases_double_positive).reshape([len(radii), len(feature_cells)])
                dnase_means = np.mean(dnases_agreement_reshaped, axis = 0)

                ######### reorder cells by similarity ################
                ## This was added 5/30/2019. It seems to *maybe help 
                ## a little bit on cell types not seen in the model.
                ## This makes sense because cell types are now ordered
                ## by similarity and keep some spacial positioning 
                ## based on the similarity to the new cell. 
                best_indices = (-dnase_means).argsort()

                dnases.extend(dnases_double_positive_reshaped[:,best_indices].flatten())
                dnases.extend(dnases_agreement_reshaped[:,best_indices].flatten())

                feature_cell_indices = feature_cell_indices.reshape([len(feature_cells), len(assaymap)])[best_indices,:].flatten()
                ######## End reorder #################################                                                   
                                                                        
                                                                        
                # Extract features
                features = data[feature_cell_indices,i]
                
                # concatenate features and DNases
                x_data = np.concatenate([features, dnases])
                
                # mask for x_data. 0 = do not mask, 1 = mask.
                x_mask = np.zeros(x_data.shape[0])
                x_mask[np.where(feature_cell_indices == -1)[0]] = True # assign mask to missing features

                # There can be NaNs in the DNases for edge cases (when the radii extends past the end of the chr).
                # Mask these values in the first row of tmp
                x_mask[np.where(np.isnan(x_data))[0]] = True # assign mask to missing DNase values
                x_data[np.where(x_mask == True)[0]] = 0 # set all UNKs to 0
                
                # mask x data by which factors are available for a cell type
                x_data_masked = np.vstack([x_mask, x_data]) # top row 0 = mask, bottom row 1 = data
                np.random.shuffle(x_data_masked )
                        
                if (mode != Dataset.RUNTIME):
                    labels = data[label_cell_indices_no_dnase,i]

                else: # used when just predicting
                    # The features going into the example.
                    labels = garbage_labels # all 0's
                    
                yield (x_data_masked, labels, assay_mask)

    return g


def generator_to_tf_dataset(g, batch_size, shuffle_size, prefetch_size):
    """
    Generates a tensorflow dataset from a data generator.
    
    :param g: data generator
    :param batch_size: number of elements in generator to combine into a single batch
    :param shuffle_size: number of elements from the  generator fromw which the new dataset will shuffle
    :param prefetch_size: maximum number of elements that will be buffered  when prefetching
    
    :returns: tuple of (label shape, tf.data.Dataset)
    """
    
    for x, y, y_mask in g():
        break
    
    try: 
        dataset = tf.data.Dataset.from_generator(
            g,
            output_types=(tf.float32,)*3, # 3 = features, labels, and missing indices
            output_shapes=(x.shape, y.shape, y.shape,)
        )
    except NameError as e:
        print("Error: no data, %s" % e)
        dataset = tf.data.Dataset.from_generator(
            g,
            output_types=(tf.float32,)*3 # 3 = features, labels, and missing indices
        )

    dataset = dataset.batch(batch_size)
    dataset = dataset.shuffle(shuffle_size)
    dataset = dataset.repeat()
    dataset = dataset.prefetch(prefetch_size)
    
    try: 
        y
        return y.shape, dataset
    except NameError as e:
        return None, dataset



import numpy as np
import tensorflow as tf
import threading


"""
Lightweight library to read and write TF-Records.
"""

################################# TF-Record Utilities ################################

def make_bytes_feature(value_list):
  return tf.train.Feature(
    bytes_list=tf.train.BytesList(value=value_list))

def make_int64_feature(value_list):
  return tf.train.Feature(
    int64_list=tf.train.Int64List(value=value_list))

def make_float_feature(value_list):
  return tf.train.Feature(
    float_list=tf.train.FloatList(value=value_list))

def make_example(features_dict):
  return tf.train.Example(
    features=tf.train.Features(feature=features_dict))

def get_bytes_list(example, key):
  return example.features.feature[key].bytes_list.value

def get_int64_list(example, key):
  return example.features.feature[key].int64_list.value

def get_float_list(example, key):
  return example.features.feature[key].float_list.value

def read_tfrecord(path, proto=None, options=None):
  if proto is None:
    proto = tf.train.Example
  string_iterator = tf.python_io.tf_record_iterator(path, options)
  for string in string_iterator:
    yield proto.FromString(string)

def write_tfrecord(protos, path, options=None):
  writer = tf.python_io.TFRecordWriter(path, options)
  for proto in protos:
    writer.write(proto.SerializeToString())
  writer.close()


def write_tfrecords(examples, filename, num_shards=100, options = None):

    filenames = sharded_filenames(filename, num_shards)
    func = lambda file: tf.python_io.TFRecordWriter(file, options)
    writers = list(map(func, filenames))
    
    for i, example in enumerate(examples):
        t = threading.Thread(
            target=writers[i % num_shards].write,
            args=(example.SerializeToString(),))
        t.daemon = True
        t.start()
        t.join()


  ################################# Sharded Filespec ################################

def sharded_filename(filename, shard, num_shards):
  format_ = lambda i: str(i).rjust(5, '0')
  return '%s-%s-%s' % (filename, format_(shard), format_(num_shards))

def sharded_filenames(filename, num_shards):
  return [sharded_filename(filename, i, num_shards) for i in range(num_shards)]


r"""
==============
epitome Module
==============
.. currentmodule:: epitome

epitome is a computational model for predicting ChIP-seq peaks in a new cell type
from chromatin accessibility and known ChIP-seq peaks from ENCODE. This module
also includes scripts for processing ENCODE peaks.

.. automodule:: epitome.models
.. automodule:: epitome.functions
.. automodule:: epitome.viz
.. automodule:: epitome.constants

"""
import os
from os.path import expanduser

__path__ = __import__('pkgutil').extend_path(__path__, __name__)

S3_DATA_PATH = 'https://epitome-data.s3-us-west-1.amazonaws.com/data.zip'


def GET_EPITOME_USER_PATH():
    return os.path.join(os.path.expanduser('~'), '.epitome')

# relative path to data
def GET_DATA_PATH():
    return os.path.join(GET_EPITOME_USER_PATH(),'data')

POSITIONS_FILE = "all.pos.bed.gz"
FEATURE_NAME_FILE = "feature_name"
REQUIRED_FILES = [POSITIONS_FILE,"train.npz","valid.npz", FEATURE_NAME_FILE,"test.npz"]



r"""
=========
Constants
=========
.. currentmodule:: epitome.constants

.. autosummary::
  :toctree: _generate/

  Dataset
"""

# imports
from enum import Enum
import numpy as np
import os

######################################################
################### CONSTANTS ########################
######################################################
class Features(Enum):
    MASK_IDX = 0
    FEATURE_IDX = 1

class Label(Enum):
    IMPUTED_UNBOUND = -3
    IMPUTED_BOUND = -2
    UNK = -1
    UNBOUND = 0
    BOUND = 1

class Dataset(Enum):
    """ Enumeration determining train, valid, test or runtime.
    """


    TRAIN = 1    # TRAINING
    r"""
    Training mode: Allows subsampling of 0s.
    """


    VALID = 2    # VALIDATION during training
    r"""
    Validation mode: Allows validation on training cell types.
    """


    TEST = 3     # TEST held out test set
    r"""
    Test mode: Disables subsampling of 0s.
    """

    RUNTIME = 4  # Using the model at runtime. No truth, just predictions
    r"""
    Runtime mode: Allows a new cell type to be predicted on.
    """



from setuptools import find_packages, setup
from version import version as this_version

try: # for pip >= 10
    from pip._internal.req import parse_requirements
except ImportError: # for pip <= 9.0.3
    from pip.req import parse_requirements

import os
import sys
import subprocess

# Utility function to read the README file.
# Used for the long_description.
def read(fname):
    return open(os.path.join(os.path.dirname(__file__), fname)).read()

# parse_requirements() returns generator of pip.req.InstallRequirement objects
install_reqs = parse_requirements('requirements.txt', session='hack')

reqs = [str(ir.req) for ir in install_reqs]

# append tensorflow or tensorflow-gpu to reqs
# need nightly build to work with tensorflow probability
TENSORFLOW_VERSION = "2.0.0.dev20190729"

try:
    subprocess.check_output(["nvidia-smi", "-L"])
    tf_req = "tf-nightly-gpu-2.0-preview==%s" % TENSORFLOW_VERSION
except:
    tf_req = "tf-nightly-2.0-preview==%s" % TENSORFLOW_VERSION

reqs.append(tf_req)

with open("README.md", "r") as fh:
    long_description = fh.read()

setup(
    name='epitome',
    version=this_version,
    description='ML model for predicting ChIP-seq peaks in new cell types from ENCODE cell lines',
    author='Alyssa Kramer Morrow',
    author_email='akmorrow@berkeley.edu',
    project_urls={
        'Documentation': 'https://readthedocs.org', # TODO
        'Source': 'https://github.com/akmorrow13/epitome'
    },
    classifiers=[
        'Topic :: Scientific/Engineering :: Bio-Informatics',
        # Pick your license as you wish (should match "license" above)
         'License :: OSI Approved :: Apache Software License',
        # Python versions supported
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3.8'
    ],
    license="Apache License, Version 2.0",
    keywords='ENCODE ChIP-seq_peaks prediction histone transcription_factor',
    install_requires=reqs,
    long_description=long_description,
    long_description_content_type='text/markdown',
    packages=find_packages(exclude=['*.test.*']),
    python_requires='>=3'
)
 



#####################################################################################
## Metric functions for evaluating model.
#####################################################################################

import sklearn.metrics
import numpy as np
import tensorflow as tf

def gini(actual, pred, sample_weight):                                                 
    # sort by preds
    df = tf.stack([actual,pred], axis = 0)
    df = tf.gather(df, tf.argsort(pred, direction='DESCENDING'), axis=1)
    linsp = tf.divide(tf.range(1,df.shape[1]+1), df.shape[1])
    linsp = tf.cast(linsp, tf.float32)
    # sum of actual
    totalPos = tf.math.reduce_sum(actual)
    cumPosFound = tf.math.cumsum(df[0])
    Lorentz = tf.divide(cumPosFound,totalPos)
    Gini = Lorentz - linsp
    return tf.reduce_sum(tf.boolean_mask(Gini, sample_weight)).numpy()

def gini_normalized(actual, pred, sample_weight = None):
    normalized_gini = gini(actual, pred, sample_weight)/gini(actual, actual, sample_weight)
    return normalized_gini


def get_performance(assaymap, preds, truth, sample_weight):


    assert(preds.shape == truth.shape)
    assert(preds.shape == sample_weight.shape)

    inv_assaymap = {v: k for k, v in assaymap.items()}

    evaluated_assays = {}

    for j in range(preds.shape[1]): # for all assays
        # sample_weight mask can only work on 1 row at a time.
        # If a given assay is not available for evaluation, sample_weights will all be 0
        # and the resulting roc_auc_score will be NaN.

        try:
            roc_score = sklearn.metrics.roc_auc_score(truth[:,j],
                                                      preds[:,j],
                                                      sample_weight = sample_weight[:, j],
                                                      average='macro')


        except ValueError:
            roc_score = np.NaN

        try:
            pr_score = sklearn.metrics.average_precision_score(truth[:,j],
                                                               preds[:,j],
                                                               sample_weight = sample_weight[:, j])


        except ValueError:
            pr_score = np.NaN

        try:
            gini_score = gini_normalized(truth[:,j],
                                               preds[:,j],
                                               sample_weight = sample_weight[:, j])

        except ValueError:
            gini_score = np.NaN
        evaluated_assays[inv_assaymap[j+1]] = {"AUC": roc_score, "auPRC": pr_score, "GINI": gini_score }

    return evaluated_assays



r"""
================
Helper functions
================
.. currentmodule:: epitome.functions

.. autosummary::
  :toctree: _generate/

  load_epitome_data
  load_bed_regions
  get_assays_from_feature_file
  bed2Pyranges
  bedtools_intersect
  bedFile2Vector
  range_for_contigs
  calculate_epitome_regions
  concatenate_all_data
"""

# imports
from epitome import *
import h5py
from scipy.io import savemat

import pandas as pd
import collections
import numpy as np
import os
from collections import Counter
from itertools import groupby
from scipy.io import loadmat
from .constants import *
import scipy.sparse
import pyranges as pr

from operator import itemgetter
import gzip
import urllib
import os
import sys
import requests
import urllib
import tqdm
from zipfile import ZipFile


# to load in positions file
import multiprocessing

################### CLASSES ##########################
# TODO move to separate file
class Region:
    """Genomic Region"""

    def __init__(self, chrom, start, end):
        self.chrom = chrom
        self.start = start
        self.end = end

    def __str__(self):
        return("%s:%d-%d" % (self.chrom, self.start, self.end))

    def __eq__(self, other):
        """Overrides the default implementation"""
        if isinstance(other, Region):
            return self.chrom == other.chrom and self.start == other.start and self.end == other.end
        return False

    def overlaps(self, other, min_bp = 1):
        #assert(type(other) == Region), "overlaps analysis must take Region as input but got %s" % type(other)
        return(other.chrom == self.chrom and Region.overlap([self.start, self.end], [other.start, other.end]) > min_bp)

    @staticmethod
    def overlap(interval1, interval2):
        """
        Computes overlap between two intervals
        """
        if interval2[0] <= interval1[0] <= interval2[1]:
            start = interval1[0]
        elif interval1[0] <= interval2[0] <= interval1[1]:
            start = interval2[0]
        else:
            return 0

        if interval2[0] <= interval1[1] <= interval2[1]:
            end = interval1[1]
        elif interval1[0] <= interval2[1] <= interval1[1]:
            end = interval2[1]
        else:
            return 0

        return(end - start)

    def greaterThan(self, other):
        assert(type(other) == Region), "overlaps analysis must take Region as input but got %s" % type(other)
        return(self.chrom > other.chrom or ( self.chrom == other.chrom and self.start > other.end ))


################### FUNCTIONS ########################
######################################################


################## HELPER FUNCTIONS ##################
def get_y_indices_for_cell(matrix, cellmap, cell):
    """
    Gets indices for a cell.

    :param matrix: cell type by assay matrix with locations
    label space.
    :param cellmap: map of celltype to iloc row in matrix
    :param cell: str cell name

    :return locations of indices for the cell name specified
    """

    return np.copy(matrix[cellmap[cell]])


def get_y_indices_for_assay(arrays, assaymap, assay):
    """
    Gets indices for a assay.

    :param matrix: cell type by assay matrix with locations
    label space.
    :param cellmap: map of celltype to iloc row in matrix
    :param cell: str cell name

    :return locations of indices for the cell name specified
    """
    # get  column for this assay
    matrix = output = np.array(arrays)
    return np.copy(matrix[:,assaymap[assay]])

def get_missing_indices_for_cell(matrix, cellmap, cell):
    """
    Gets indices of missing factors for a given cell type.

    :param matrix: cell type by assay matrix with locations
    label space.
    :param cellmap: map of celltype to iloc row in matrix
    :param cell: str cell name

    :return indices where this cell has no data (==-1)
    """

    indices = get_y_indices_for_cell(matrix, cellmap, cell)
    return np.where(indices == -1)[0]


################## LOADING DATA ######################

def download_and_unzip(url, dst):
    """ Downloads a url to local destination, unzips it and deletes zip.
    
    Args:
        :param url: url to download.
        :param dst: local absolute path to download data to.
    """
    dst = os.path.join(dst, os.path.basename(url))
    
    # download data if it does not exist
    if not os.path.exists(dst):

        file_size = int(urllib.request.urlopen(url).info().get('Content-Length', -1))
        if os.path.exists(dst):
            first_byte = os.path.getsize(dst)
        else:
            first_byte = 0
        if first_byte < file_size:

            header = {"Range": "bytes=%s-%s" % (first_byte, file_size)}
            pbar = tqdm.tqdm(
                total=file_size, initial=first_byte,
                unit='B', unit_scale=True, desc="Dataset not found. Downloading Epitome data to %s..." % dst)
            req = requests.get(url, headers=header, stream=True)
            with(open(dst, 'ab')) as f:
                for chunk in req.iter_content(chunk_size=1024):
                    if chunk:
                        f.write(chunk)
                        pbar.update(1024)
            pbar.close()

    if url.endswith('.zip'):
        
        # Extract zip data if it does not exist
        if not os.path.exists(dst.split('.zip')[0]):
            with ZipFile(dst, 'r') as zipObj:
               zipObj.extractall(os.path.dirname(dst))
            
            # delete old zip to free space
            os.remove(dst)
        
def load_epitome_data(data_dir=None):
    """
    Loads data processed using data/download_encode.py. This will load three sparse matrix files
    (.npz). One for train, valid (chr7) and test (chr 8 and 9). If data is not available,
    downloads Epitome data from S3.


    Args:
        :param data_dir: Directory containing train.npz, valid.npz, test.npz,
        all.pos.bed.gz file and feature_name files saved by data/download_encode.py script.
        Defaults to data in module.

    :returns: train_data, valid_data, and test_data
        3 numpy ndarrays for train, valid and test
    """
    
    if not data_dir:
        data_dir = GET_DATA_PATH()
        download_and_unzip(S3_DATA_PATH, GET_EPITOME_USER_PATH()) 

    # make sure all required files exist
    required_paths = [os.path.join(data_dir, x) for x in REQUIRED_FILES]
    assert(np.all([os.path.exists(x) for x in required_paths]))
    npz_files = list(filter(lambda x: x.endswith(".npz"), required_paths))

    sparse_matrices = [scipy.sparse.load_npz(x).toarray() for x in npz_files]
    return {Dataset.TRAIN: sparse_matrices[0], Dataset.VALID: sparse_matrices[1], Dataset.TEST: sparse_matrices[2]}

################### Parsing Deepsea Files ########################

def load_bed_regions(bedfile):
    ''' Loads Deepsea bed file (stored as .gz format), removing
    regions that have no data (the regions between valid/test
    and chromosomes X/Y).

    Args:
        :param bedfile: path to bed file.

    Returns:
        :return list of genomic Regions the size of train/valid/test data.
    '''
    
    with gzip.open(bedfile, 'r') as f:
            liPositions = f.readlines()

    def fromString(x):
        tmp = x.decode("utf-8").split('\t')
        return Region(tmp[0], int(tmp[1]), int(tmp[2]))

    return list(map(lambda x: fromString(x), liPositions))


def list_assays(feature_name_file = None):
    
    ''' Parses a feature name file from DeepSea. File can be found in repo at ../data/feature_name.
    Returns at matrix of cell type/assays which exist for a subset of cell types.

    Args:
        :param: feature_name_file. Path to file containing cell, ChIP metadata. Defaults to module data feature file.

    Returns:
        assays: list of assay names
    '''
    if not feature_name_file:
        feature_name_file = os.path.join(DATA_PATH, FEATURE_NAME_FILE)    

    with open(feature_name_file) as f:

        assays=[]    # dict of {cell: {dict of indexed assays} }
        for i,l in enumerate(f):
            if (i == 0):
                continue

            # for example, split '8988T|DNase|None'
            _, assay = l.split('\t')[1].split('|')[:2]
            assays.append(assay)

    return assays

    

def get_assays_from_feature_file(feature_name_file = None,
                                 eligible_assays = None,
                                 eligible_cells = None,
                                 min_cells_per_assay= 3,
                                 min_assays_per_cell = 2):
    ''' Parses a feature name file. File can be found in repo at ../data/feature_name.
    Returns at matrix of cell type/assays which exist for a subset of cell types.

    Args:
        :param: feature_name_file. Path to file containing cell, ChIP metadata. Defaults to module data feature file.
        :param eligible_assays: list of assays to filter by (ie ["CTCF", "EZH2", ..]). If None, then returns all assays.
        Note that DNase will always be included in the factors, as it is required by Epitome.
        :param eligible_cells: list of cells to filter by (ie ["HepG2", "GM12878", ..]). If None, then returns all cell types.
        :param min_cells_per_assay: number of cell types an assay must have to be considered
        :param min_assays_per_cell: number of assays a cell type must have to be considered. Includes DNase.

    Returns:
        matrix: cell type by assay matrix
        cellmap: index of cells
        assaymap: index of assays
    '''
    
    if not feature_name_file:
        feature_name_file = os.path.join(DATA_PATH, FEATURE_NAME_FILE)

    # check argument validity
    if (min_assays_per_cell < 2):
         print("Warning: min_assays_per_cell should not be < 2 (this means it only has DNase) but was set to %i" % min_assays_per_cell)


    if (min_cells_per_assay < 2):
         print("Warning: min_cells_per_assay should not be < 2 (this means you may only see it in test) but was set to %i" % min_cells_per_assay)

    if (eligible_assays != None):
        if (len(eligible_assays) + 1 < min_assays_per_cell):
            raise Exception("""%s is less than the minimum assays required (%i).
            Lower min_assays_per_cell to (%i) if you plan to use only %i eligible assays""" \
                            % (eligible_assays, min_assays_per_cell, len(eligible_assays)+1, len(eligible_assays)))

    if (eligible_cells != None):
        if (len(eligible_cells) + 1 < min_cells_per_assay):
            raise Exception("""%s is less than the minimum cells required (%i).
            Lower min_cells_per_assay to (%i) if you plan to use only %i eligible cells""" \
                            % (eligible_cells, min_cells_per_assay, len(eligible_cells)+1, len(eligible_cells)))


    # Want a dictionary of assay: {list of cells}
    # then filter out assays with less than min_cells_per_assay cells
    # after this, there may be some unused cells so remove those as well
    with open(feature_name_file) as f:

        indexed_assays={}    # dict of {cell: {dict of indexed assays} }
        for i,l in enumerate(f):
            if (i == 0):
                continue

            # for example, split '8988T|DNase|None'
            cell, assay = l.split('\t')[1].split('|')[:2]

            # check if cell and assay is valid
            valid_cell = (eligible_cells == None) or (cell in eligible_cells)
            valid_assay = (eligible_assays == None) or (assay in eligible_assays) or (assay == "DNase")

            # if cell and assay is valid, add it in
            if valid_cell and valid_assay:
                if cell not in indexed_assays:
                    indexed_assays[cell] = {assay: i-1} # add index of assay
                else:
                    indexed_assays[cell][assay] = i-1



    # finally filter out cell types with < min_assays_per_cell and have DNase
    indexed_assays = {k: v for k, v in indexed_assays.items() if 'DNase' in v.keys() and len(v) >= min_assays_per_cell}

    # make flatten list of assays from cells
    tmp = [list(v) for k, v in indexed_assays.items()]
    tmp = [item for sublist in tmp for item in sublist]

    # list of assays that meet min_cell criteria
    valid_assays = {k:v for k, v in Counter(tmp).items() if v >= min_cells_per_assay}

    # remove invalid assays from indexed_assays
    for key, values in indexed_assays.items():

        # remove assays that do not mean min_cell criteria
        new_v = {k: v for k, v in values.items() if k in valid_assays.keys()}
        indexed_assays[key] = new_v

    potential_assays = valid_assays.keys()
    cells = indexed_assays.keys()

    # sort cells alphabetical
    cells = sorted(cells, reverse=True)

    # sort assays alphabetically
    potential_assays = sorted(potential_assays, reverse=True)

    # make sure DNase is first assay. This is because the model
    # assumes the first column specifies DNase
    potential_assays.remove("DNase")
    potential_assays.insert(0,"DNase")

    cellmap = {cell: i for i, cell in enumerate(cells)}
    assaymap = {assay: i for i, assay in enumerate(potential_assays)}

    matrix = np.zeros((len(cellmap), len(assaymap))) - 1
    for cell in cells:
        for assay, _ in indexed_assays[cell].items():
            matrix[cellmap[cell], assaymap[assay]] = indexed_assays[cell][assay]

    matrix = matrix.astype(int)
    return matrix, cellmap, assaymap


################### Parsing data from bed file ########################


def bed2Pyranges(bed_file):
    """
    Loads bed file in as a pyranges object.
    Preserves ordering of bed lines.

    Args:
        :param bed_file: absolute path to bed file
    Returns:
        indexed pyranges object
    """
    # just get chromosome location (first three columns)
    p = pd.read_csv(bed_file, sep='\t',header=None)[[0,1,2]]

    p['idx']=p.index
    p.columns = ['Chromosome', 'Start','End','idx']
    return pr.PyRanges(p).sort()


def bedtools_intersect(file_triple):
    """
    Runs intersection between 2 bed files and returns a vector of 0/1s
    indicating absense or presense of overlap.

    Args:
        :param file_triple: triple of (bed_file_1, bed_file_2, boolean).
            bed_file_1: bed file to run intersection against.
            bed_file_2: bed file to check for overlaps with bed_file_1.
            boolean: boolean determines wheather to return
            original peaks from bed_file_1.

    Returns:
        tuple of (bed_file_1 peaks, vector of 0/1s) whose length is len(bed_file_1).
        1s in vector indicate overlap of bed_file_1 and bed_file_2).

    """
    bed1 = bed2Pyranges(file_triple[0])
    bed2 = bed2Pyranges(file_triple[1])

    res = bed1.join(bed2, how='left')
    overlap_vector = np.zeros(len(bed1),dtype=bool)

    # get regions with overlap and set to 1
    res_df = res.df
    if not res_df.empty: # throws error if empty because no columns
        overlap_vector[res_df[res_df['Start_b'] != -1]['idx']] = 1

    if (file_triple[2]):
        # for some reason chaining takes a lot longer, so we run ops separately.
        t1 = bed1.df.sort_values(by='idx')[['Chromosome','Start','End']]
        t2 = t1.values
        t3 = t2.tolist()
        return (list(map(lambda x: Region(x[0],x[1],x[2]), t3)), overlap_vector)
    else:
        return (None, overlap_vector)

def bedFile2Vector(bed_file, allpos_bed_file):
    """
    This function takes in a bed file of peaks and converts it to a vector or 0/1s that can be
    uses as input into an Epitome model. Each 0/1 represents a region in the train/test/validation set from DeepSEA.

    Most likely, the bed file will be the output of the IDR function, which detects peaks based on the
    reproducibility of multiple samples.

    Args:
        :param bed_file: bed file containing peaks
        :param allpos_bed_file: bed file containing all positions in the dataset

    Returns:
        :return: tuple (numpy_train_array, (bed_peaks, numpy_bed_array).
        numpy_train_array: boolean numpy array indicating overlap of training data with peak file (length of training data).
        bed_peaks: a list of intervals loaded from bed_file.
        numpy_bed_array: boolean numpy array indicating presence or absense of each bed_peak region in the training dataset.
    """

    bed_files = [(allpos_bed_file, bed_file, False), (bed_file, allpos_bed_file, True)]
    pool = multiprocessing.Pool(processes=2)
    results = pool.map(bedtools_intersect, bed_files)

    pool.close()
    pool.join()

    return (results[0][1], results[1])

def indices_for_weighted_resample(data, n,  matrix, cellmap, assaymap, weights = None):
    """
    Selects n rows from data that have the greatest number of labels (can be weighted)
    Returns indices to these rows.

    :param data: data matrix with shape (factors, records)
    :param n: number or rows to sample
    :param matrix: cell type by assay position matrix
    :param cellmap dict of cells and row positions in matrix
    :param assaymap: dict of assays and column positions in matrix
    :param weights: Optional vector of weights whos length = # factors (1 weight for each factor).
    The greater the weight, the more the positives for this factor matters.
    """

    # only take rows that will be used in set
    # drop DNase from indices in assaymap first
    selected_assays = list(assaymap.values())[1:]
    indices = matrix[list(cellmap.values())][:,selected_assays].flatten()

    # set missing assay/cell combinations to -1
    t1 = data[indices, :]
    t1[np.where(indices < 0)[0],:] = 0

    # sum over each factor for each record
    sums = np.sum(np.reshape(t1, (len(selected_assays), len(cellmap), t1.shape[1])), axis=1)

    if (weights is not None):
        weights = np.reshape(weights, (weights.shape[0],1)) # reshape so multiply works
        probs = np.sum(sums * weights, axis = 0)
        probs = probs/np.sum(probs)
    else:
        # simple sum over recoreds. Weights records with more positive
        # samples higher for random sampling.
        probs = np.sum(sums, axis=0)
        probs = (probs)/np.sum(probs)

    # TODO assign equal probs to non-zero weights
    probs[probs != 0] = 1/probs[probs != 0].shape[0]

    radius = 20

    n = int(n / radius)
    data_count = data.shape[1]

    # sample by probabilities. not sorted.
    choice = np.random.choice(np.arange(0, data_count), n, p = probs)

    func_ = lambda x: np.arange(x - radius/2, x + radius/2)
    surrounding = np.unique(list(map(func_, choice)))
    return surrounding[(surrounding > 0) & (surrounding < data_count)].astype(int)


##################### Functions for parsing allpos.bed file ################
def range_for_contigs(all_regions_file):
    """
    Traverses through feature_name_file to get contig ranges.

    Args:
        :param all_regions_file: path to bed file of genomic regions

    Returns:
        list of contigs and their start/end position all_regions_file
    """
    with gzip.open(all_regions_file,'rt') as f:

        contigs = {}
        this_contig = "placeholder"
        this_range = [0,0]
        for num, line in enumerate(f):
            contig = line.split("\t")[0]

            if (this_contig != contig):
                # update last contig and save information
                this_range[1] = num
                contigs[this_contig] = this_range

                # reset contig to new
                this_contig = contig
                this_range = [0,0]
                this_range[0] = num
            else:
                continue

        # add last contig
        this_range[1] = num + 1
        contigs[this_contig] = this_range

        del contigs["placeholder"]

        return contigs

def calculate_epitome_regions(all_regions_file):
    """
    Gets line numbers for train/valid/test boundaries.
    Assumes that chr7,8,9 are all in a row (only true if sex chrs are removed).


    Args:
        :param all_pos_file: bed file containing a row for each genomic region.

    Returns:
        triple of train,valid,test regions
    """

    contig_ranges = range_for_contigs(all_regions_file)

    EPITOME_VALID_REGIONS = contig_ranges["chr7"]
    # 227512 for test (chr8 and chr9)
    chr8_start = contig_ranges["chr8"][0]
    chr9_end = contig_ranges["chr9"][1]
    EPITOME_TEST_REGIONS  = [chr8_start, chr9_end] # chr 8 and 9

    EPITOME_TRAIN_REGIONS = [[0,EPITOME_VALID_REGIONS[0]],
                             [EPITOME_TEST_REGIONS[1],contig_ranges["chr21"][1]]]


    return(EPITOME_TRAIN_REGIONS, EPITOME_VALID_REGIONS, EPITOME_TEST_REGIONS)


def concatenate_all_data(data, region_file):
    """
    Puts data back in correct order to be indexed by the allpos file.

    Args:
        :param data: data dictionary of train, valid and test
        :param regions_file: bed file containg regions of train, valid and test.

    Returns:
        np matrix of concatenated data
    """

    # Get chr6 cutoff. takes about 3s.
    chr6_end = range_for_contigs(region_file)['chr6'][1]
    return np.concatenate([data[Dataset.TRAIN][:,0:chr6_end], # chr 1-6, range_for_contigs is 1 based
                           data[Dataset.VALID], # chr7
                           data[Dataset.TEST], # chr 8 and 9
                           data[Dataset.TRAIN][:,chr6_end:]],axis=1) # all the rest of the chromosomes



from epitome.test import EpitomeTestCase
from epitome.constants import Dataset

class ModelsTest(EpitomeTestCase):

	def test_trainModel(self):
		train_iters = 50
		validation_size = 10
		
		# create model and train
		model = self.makeSmallModel()
		model.train(train_iters)
		results = model.test(validation_size)

		# Make sure predictions are not random
		# after first iterations
		assert(results['preds_mean'].numpy()[0,0] < 0.1)
		assert(results['preds_mean'].shape[0] == validation_size)


		# make sure can run in test mode
		results = model.test(validation_size, mode=Dataset.TEST)
		assert(results['preds_mean'].numpy()[0,0] < 0.1)
